/* cache.c   simulate cache design

   P. Chongstitvatana	8 Sept 97 
*/

#include <stdio.h>
#include <stdlib.h>

#define		INT	int
#define		MAX	10000

void init();
void replace(int idx, int tn);
int readtrace(FILE *fn);
int leastrecent(int idx);
void updateLRU(int n, int idx);

int  W,CS,M,H,LRU;
int  tag[4][MAX];
char lru[4][MAX];
FILE *fn;
INT hit, total;

int linesize[] = {1,2,4,8};
int way[] = {1,2,4};
int cachesize[] = {128,256,512,1024,2048,4096,8192}; 

double ratio;

main()
{
  int i,j,k;
  LRU = 1;
  for( k=0; k<4; k++) {
    CS = cachesize[k];
    for( j=0; j<3; j++) {
      M = way[j];
      printf("CS %d M %d\n",CS,M);
      for( i=0; i<4; i++ ) {
	W = linesize[i];
	trace();
	ratio = (double)hit / (double) total;
	printf("W %d H %d hit ratio %5.3f\n",W,H,ratio);
      }
    }
  }
}

trace()
{
  int a,idx,tn,i,m;
  init();
  fn = fopen("trace.tr","rt");
  while( (a = readtrace(fn)) != EOF ) {
/*    if ( total > 10000 ) break;   */
    total++;
    a = a / W;
    idx = a % H;
    tn = a / H;
    for( i=0; i<M; i++ )
      if( tag[i][idx] == tn ) {  /* hit */
	hit++;
	updateLRU(i,idx);
	break;
      }
    if( i == M ) replace(idx,tn); /* miss */
  }
  fclose(fn);
}

void replace( int idx, int tn )
{
  int m;
  if ( LRU ) 
    m = leastrecent(idx);       
  else 
    m = lrand48() % M;
  tag[m][idx] = tn;
  updateLRU(m,idx);
}   

int leastrecent(int idx)
{
  int i;
  for( i=0; i<M; i++ )
    if( lru[i][idx] == 0 ) break;
  return i;
}

void updateLRU( int n, int idx )
{
  int c,i;
  c = lru[n][idx];
  for( i=0; i<M; i++ )
    if( lru[i][idx] > c ) lru[i][idx]--;
  lru[n][idx] = M-1;
}

int readtrace(FILE *fn)
{
  int a;
  if( fscanf(fn,"%d\n",&a) == EOF ) return EOF;
  return a;
}

void init()
{
  int i,j;
  hit = 0;
  total = 0;
  H = CS/(M*W);
  for( j=0; j<M; j++ )
    for( i=0; i<CS; i++ ) {
      tag[j][i] = 0;
      lru[j][i] = j;
    }
}

